/* simpleEvent_Tk.c --
 *
 * This file is just a way to get simpleEvent compiled for use with Tk.
 */

/* $Header: /cluster21/kennykb/src/tclTCP.1.0beta/RCS/simpleEvent_Tk.c,v 1.1 1992/02/20 16:22:09 kennykb Exp $
 * $Source: /cluster21/kennykb/src/tclTCP.1.0beta/RCS/simpleEvent_Tk.c,v $
 * $Log: simpleEvent_Tk.c,v $
 * Revision 1.1  1992/02/20  16:22:09  kennykb
 * Initial revision
 *
 */

#define USE_TK
#define USE_SIGNALS 0
#include "simpleEvent.c"

